/*
Martrikelnummer der Gruppenmitglieder: Olha Grubryn (235122), Egor Rutkowski(225064), Nikola Kramaric (238498), Dennis Ruppel (238503)
*/
package com.example;

import akka.actor.typed.ActorRef;
import akka.actor.typed.Behavior;
import akka.actor.typed.javadsl.AbstractBehavior;
import akka.actor.typed.javadsl.ActorContext;
import akka.actor.typed.javadsl.Behaviors;
import akka.actor.typed.javadsl.Receive;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Random;

/**
 * AddressBook Actor class that maintains a list of customers and handles messages related to customer addresses.
 * This actor can respond to requests for customer addresses and can provide the list of all customers.
 */
public class AddressBook extends AbstractBehavior<AddressBook.Message> {

    private final ArrayList<ActorRef<Customer.Message>> customers;

    /**
     * Interface for all messages that AddressBook can handle.
     */
    public interface Message {}

    /**
     * Message sent by a customer asking for a random address from the AddressBook.
     */
    public static final class CustomerMessage implements Message {
        public final ActorRef<Customer.Message> fragender;  // The customer actor sending the message
        public final String nameFragender;  // The name of the customer
        public final ActorRef<DeliveryCar.Message> truckReference;  // Reference to the delivery car

        public CustomerMessage(ActorRef<Customer.Message> fragender, String nameFragender, ActorRef<DeliveryCar.Message> truckReference) {
            this.fragender = fragender;
            this.nameFragender = nameFragender;
            this.truckReference = truckReference;
        }
    }

    /**
     * Message to request the list of all customers.
     */
    public static final class GetCustomers implements Message {
        public final ActorRef<List<ActorRef<Customer.Message>>> replyTo;  // Actor to reply to with the list of customers

        public GetCustomers(ActorRef<List<ActorRef<Customer.Message>>> replyTo) {
            this.replyTo = replyTo;
        }
    }

    /**
     * Constructor for AddressBook.
     * Initializes the AddressBook with a list of customer actors and sends each customer an AddressHello message.
     *
     * @param context the actor context
     * @param customers the list of customer actors
     */
    private AddressBook(ActorContext<Message> context, ArrayList<ActorRef<Customer.Message>> customers) {
        super(context);
        this.customers = new ArrayList<>(customers);
        // Notify each customer with an AddressHello message
        for (ActorRef<Customer.Message> cst : this.customers) {
            cst.tell(new Customer.AddressHello(getContext().getSelf()));
        }
    }

    /**
     * Factory method to create a new AddressBook actor.
     *
     * @param customers the list of customer actors
     * @return the behavior of the AddressBook actor
     */
    public static Behavior<Message> create(ArrayList<ActorRef<Customer.Message>> customers) {
        return Behaviors.setup(context -> new AddressBook(context, customers));
    }

    @Override
    public Receive<Message> createReceive() {
        return newReceiveBuilder()
                .onMessage(CustomerMessage.class, this::onCustomerMessage)
                .onMessage(GetCustomers.class, this::onGetCustomers)
                .build();
    }

    /**
     * Handles CustomerMessage by selecting a random customer from the list and sending their address to the requesting customer.
     *
     * @param msg the customer message requesting a random address
     * @return the behavior of the AddressBook actor
     */
    private Behavior<Message> onCustomerMessage(CustomerMessage msg) {
        Random random = new Random();
        int index = random.nextInt(customers.size());
        ActorRef<Customer.Message> addressOf = customers.get(index);
        getContext().getLog().info("{} hat nach einer Adresse gefragt CAR: {}", msg.nameFragender, msg.truckReference.path().name());
        msg.fragender.tell(new Customer.AddressMessage(addressOf, msg.truckReference));
        return this;
    }

    /**
     * Handles GetCustomers message by replying with the unmodifiable list of customers.
     *
     * @param msg the get customers message
     * @return the behavior of the AddressBook actor
     */
    private Behavior<Message> onGetCustomers(GetCustomers msg) {
        msg.replyTo.tell(Collections.unmodifiableList(customers));
        return this;
    }
}
