/*
Martrikelnummer der Gruppenmitglieder: Olha Grubryn (235122), Egor Rutkowski(225064), Nikola Kramaric (238498), Dennis Ruppel (238503)
*/
package com.example;

import akka.actor.typed.ActorRef;
import akka.actor.typed.Behavior;
import akka.actor.typed.javadsl.AbstractBehavior;
import akka.actor.typed.javadsl.ActorContext;
import akka.actor.typed.javadsl.Behaviors;
import akka.actor.typed.javadsl.Receive;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Random;

/**
 * Customer Actor class that simulates a customer in the system.
 * This actor can receive packages, send packages, and interact with an address book.
 */
public class Customer extends AbstractBehavior<Customer.Message> {

    String name;
    ActorRef<AddressBook.Message> adrBook;
    ArrayList<String> gegenstaende = new ArrayList<>(Arrays.asList("Spülmaschine","Altes CD-Regal", "Goldbarren","20kg Hanteln", "Holzkohlegrill","Blumenerde"));

    /**
     * Interface for all messages that Customer can handle.
     */
    public interface Message {}

    /**
     * Message representing a delivery of a package to the customer.
     */
    public static final class DeliveryMessage implements Message {
        public final Paket paket;

        public DeliveryMessage(Paket paket) {
            this.paket = paket;
        }
    }

    /**
     * Message representing a request to pick up a package from the customer.
     */
    public static final class PickUpMessage implements Message {
        public final ActorRef<DeliveryCar.Message> truckReference;
        public final String someString;

        public PickUpMessage(ActorRef<DeliveryCar.Message> truckReference, String someString) {
            this.truckReference = truckReference;
            this.someString = someString;
        }
    }

    public record AddressHello(ActorRef<AddressBook.Message> book) implements Message {}
    public record AddressMessage(ActorRef<Customer.Message> addresse, ActorRef<DeliveryCar.Message> truckReference) implements Message {}

    /**
     * Constructor for Customer.
     * Initializes the customer with a name.
     *
     * @param context the actor context
     * @param name the name of the customer
     */
    private Customer(ActorContext<Message> context, String name){
        super(context);
        this.name = name;
    }


    /**
     * Factory method to create a new Customer actor.
     *
     * @param name the name of the customer
     * @return the behavior of the Customer actor
     */
    public static Behavior<Customer.Message> create(String name) {
        return Behaviors.setup(context -> new Customer(context, name));
    }

    //Verarbeiten von den DeliveryMessage und PickUpMessage Nachrichten
    @Override
    public Receive<Message> createReceive() {
        return newReceiveBuilder()
                .onMessage(DeliveryMessage.class, this::onDeliveryMessage)
                .onMessage(AddressHello.class, this::onAddressHello)
                .onMessage(PickUpMessage.class, this::onPickUpMessage)
                .onMessage(AddressMessage.class, this::onAddressMessage)
                .build();
    }

    /**
     * Stores the address book reference when received.
     *
     * @param msg the AddressHello message
     * @return the behavior of the Customer actor
     */
    private Behavior<Message> onAddressHello(AddressHello msg){
        this.adrBook = msg.book;
        return this;
    }

    /**
     * Handles the PickUpMessage by deciding whether to send a package.
     * If a random number is less than or equal to 80, a request is sent to the address book
     * to get a random customer address, and the package is then sent.
     * Otherwise, logs that no package will be sent.
     *
     * @param msg the PickUpMessage
     * @return the behavior of the Customer actor
     */
    private Behavior<Message> onPickUpMessage(PickUpMessage msg){
        getContext().getLog().info("Kunde {}, will kein Paket abschicken CAR: {}", this.name, msg.truckReference.path().name());
        Random random = new Random();
        int number = random.nextInt(101);
        if(number <= 80){
            //truck wird hier übergeben, damit man im nächsten Schritt wieder auf ihn zugreifen kann.
            this.adrBook.tell(new AddressBook.CustomerMessage(getContext().getSelf(), this.name, msg.truckReference));

            //Methode um einen Paket zu versenden
            //über onAddressMessage wird dann Paket aufgenommen
            onAddressMessage(new AddressMessage(getContext().getSelf(), msg.truckReference));
        }else{
            getContext().getLog().info("Kunde {}, will kein Paket abschicken", this.name);
            msg.truckReference.tell(new DeliveryCar.PickupResponse(null));
        }
        return this;
    }

    /**
     * Handles the AddressMessage by creating a package with a random item and sending it to the truck.
     *
     * @param adr the AddressMessage
     * @return the behavior of the Customer actor
     */
    private Behavior<Message> onAddressMessage(AddressMessage adr){
        Random random = new Random();
        int index = random.nextInt(gegenstaende.size());
        String gegenstand = gegenstaende.get(index);
        Paket paket = new Paket(gegenstand,this.name, adr.addresse);
        getContext().getLog().info("Paket mit dem Inhalt {} wurde an {} verschickt (CAR: {})", paket.inhalt, adr.addresse.path().name(), adr.truckReference.path().name());
        adr.truckReference.tell(new DeliveryCar.PickupResponse(paket));
        return this;
    }

    /**
     * Handles the DeliveryMessage by logging the received package details.
     *
     * @param pkt the DeliveryMessage
     * @return the behavior of the Customer actor
     */
    private Behavior<Message> onDeliveryMessage(DeliveryMessage pkt){
        getContext().getLog().info("Ich habe ein Paket von {} erhalten mit dem Inhalt: {} ", pkt.paket.absender, pkt.paket.inhalt);
        return this;
    }
}