/*
Martrikelnummer der Gruppenmitglieder: Olha Grubryn (235122), Egor Rutkowski(225064), Nikola Kramaric (238498), Dennis Ruppel (238503)
*/
package com.example;

import akka.actor.typed.ActorRef;
import akka.actor.typed.Behavior;
import akka.actor.typed.javadsl.*;

import java.time.Duration;
import java.util.ArrayList;
import java.util.List;

import static com.example.AkkaMainSystem.getDistributionCenter;

/**
 * DeliveryCar class that simulates a delivery car in the system.
 * This actor can load packages, follow a delivery route, and pick up packages from customers.
 */
public class DeliveryCar extends AbstractBehavior<DeliveryCar.Message> {

    // Liste der Pakete im Lieferwagen
    private final ArrayList<Paket> pakete = new ArrayList<>();
    // Route mit Kunden, die abgefahren wird
    private final ArrayList<ActorRef<Customer.Message>> route;
    // Maximale Kapazität des Lieferwagens
    private final int maxCapacity = 3;
    // Timer zum Planen von Zustellungen
    private final TimerScheduler<Message> timers;
    // Globaler Index für die aktuelle Position auf der Route
    private int globalIndex = 0;

    /**
     * Interface for all messages that DeliveryCar can handle.
     */
    public interface Message {}

    /**
     * Message to load packages into the delivery car.
     */
    public record LoadMessage(List<Paket> pakete) implements Message {}

    /**
     * Message representing the response from a customer when a package is picked up.
     */
    public record PickupResponse(Paket paket) implements Message {}

    /**
     * Message to start the delivery route.
     */
    public record StartRoute() implements Message {}

    /**
     * Message to check the next stop on the route.
     */
    public record CheckStop() implements Message {}

    /**
     * Message to clear all packages in the delivery car.
     */
    public record Clear() implements Message {}

    /**
     * Constructor for DeliveryCar.
     *
     * @param context the actor context
     * @param timers the timer scheduler
     * @param route the delivery route
     */
    private DeliveryCar(ActorContext<Message> context, TimerScheduler<Message> timers, ArrayList<ActorRef<Customer.Message>> route) {
        super(context);
        this.route = route;
        this.timers = timers;
    }

    /**
     * Factory method to create a new DeliveryCar actor.
     *
     * @param route the delivery route
     * @return the behavior of the DeliveryCar actor
     */
    public static Behavior<Message> create(List<ActorRef<Customer.Message>> route) {
        return Behaviors.setup(context -> Behaviors.withTimers(timers -> new DeliveryCar(context, timers, new ArrayList<>(route))));
    }

    // Handling of the various messages the DeliveryCar can receive
    @Override
    public Receive<Message> createReceive() {
        return newReceiveBuilder()
                .onMessage(PickupResponse.class, this::onPickupResponse)
                .onMessage(LoadMessage.class, this::onLoadMessage)
                .onMessage(StartRoute.class, this::onStartRoute)
                .onMessage(CheckStop.class, this::onCheckStop)
                .onMessage(Clear.class, this::onClear)
                .build();
    }

    /**
     * Handles the LoadMessage to load packages into the delivery car.
     *
     * @param msg the LoadMessage
     * @return the behavior of the DeliveryCar actor
     */
    private Behavior<Message> onLoadMessage(LoadMessage msg) {
        pakete.addAll(msg.pakete);
        getContext().getLog().info("Laden: {} Pakete wurden in den LKW geladen.", msg.pakete.size());
        scheduleNextStopCheck(Duration.ofSeconds(3)); // Start checking the first stop after 3 seconds
        return this;
    }

    /**
     * Handles the PickupResponse to pick up a package from a customer.
     *
     * @param rsp the PickupResponse
     * @return the behavior of the DeliveryCar actor
     */
    private Behavior<Message> onPickupResponse(PickupResponse rsp) {
        if (rsp.paket != null && rsp.paket.inhalt != null && pakete.size() < maxCapacity) {
            pakete.add(rsp.paket);
            getContext().getLog().info("Abholung: Ein Paket wurde aufgenommen. Aktuelle Anzahl der Pakete im LKW: {}", pakete.size());
        }
        scheduleNextStopCheck(Duration.ofSeconds(1)); // Move to the next stop after 1 second
        return this;
    }

    /**
     * Checks if the delivery car has completed the route and returns to the distribution center.
     * Otherwise, it moves to the next stop.
     *
     * @param stop the CheckStop message
     * @return the behavior of the DeliveryCar actor
     */
    private Behavior<Message> onCheckStop(CheckStop stop) {
        if (globalIndex >= route.size()) {
            getContext().getLog().info("Rückkehr: Der LKW kehrt zum Verteilzentrum zurück mit {} Paketen.", pakete.size());
            DistributionCenter.ArriveMessage antwort = new DistributionCenter.ArriveMessage(pakete, getContext().getSelf());
            getDistributionCenter().tell(antwort);
            globalIndex = 0;
        } else {
            ActorRef<Customer.Message> currentStop = route.get(globalIndex);
            getContext().getLog().info("Aktuelle Haltestelle: {}", currentStop.path().name());
            deliverPackages(currentStop);
            if (pakete.size() < maxCapacity) {
                sendPickupMessage(currentStop);
            }
            globalIndex++;
            scheduleNextStopCheck(pakete.size() >= maxCapacity ? Duration.ofSeconds(1) : Duration.ofSeconds(3));
        }
        return this;
    }

    /**
     * Starts the delivery route.
     *
     * @param msg the StartRoute message
     * @return the behavior of the DeliveryCar actor
     */
    private Behavior<Message> onStartRoute(StartRoute msg) {
        globalIndex = 0;
        scheduleNextStopCheck(Duration.ofSeconds(3)); // Start checking the first stop after 3 seconds
        return this;
    }

    /**
     * Clears all packages in the delivery car.
     *
     * @param clear the Clear message
     * @return the behavior of the DeliveryCar actor
     */
    private Behavior<Message> onClear(Clear clear) {
        pakete.clear();
        return this;
    }

    /**
     * Sends a pickup message to a customer to ask if they have a package to send.
     *
     * @param customer the customer actor reference
     */
    private void sendPickupMessage(ActorRef<Customer.Message> customer) {
        customer.tell(new Customer.PickUpMessage(getContext().getSelf(), "Der LKW kann das Paket abholen."));
    }

    /**
     * Delivers packages that are meant for the current customer.
     *
     * @param customer the customer actor reference
     */
    private void deliverPackages(ActorRef<Customer.Message> customer) {
        pakete.removeIf(paket -> {
            if (paket.empfaenger.equals(customer)) {
                customer.tell(new Customer.DeliveryMessage(paket));
                return true;
            }
            return false;
        });
    }

    /**
     * Schedules the next stop check after a given duration.
     *
     * @param duration the duration after which to check the next stop
     */
    private void scheduleNextStopCheck(Duration duration) {
        timers.startSingleTimer(new CheckStop(), duration);
    }
}
