/*
Martrikelnummer der Gruppenmitglieder: Olha Grubryn (235122), Egor Rutkowski(225064), Nikola Kramaric (238498), Dennis Ruppel (238503)
*/
package com.example;

import akka.actor.typed.ActorRef;
import akka.actor.typed.Behavior;
import akka.actor.typed.javadsl.AbstractBehavior;
import akka.actor.typed.javadsl.ActorContext;
import akka.actor.typed.javadsl.Behaviors;
import akka.actor.typed.javadsl.Receive;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * Die Klasse `DistributionCenter` repräsentiert das Verteilzentrum im System.
 * Das Verteilzentrum empfängt Pakete von Lieferwagen, speichert sie und generiert Routen für Lieferwagen.
 */
public class DistributionCenter extends AbstractBehavior<DistributionCenter.Message> {
    // Referenz auf das Adressbuch
    private final ActorRef<AddressBook.Message> addressBook;
    // Liste der Pakete im Lager
    private final List<Paket> lagerraum = new ArrayList<>();
    // Liste der Lieferwagen
    private final List<ActorRef<DeliveryCar.Message>> deliveryCars = new ArrayList<>();

    public static class StartAllRoutes implements Message {}


    /**
     * Konstruktor für das Verteilzentrum.
     *
     * @param context das Actor-Kontext
     * @param addressBook Referenz auf das Adressbuch
     */
    private DistributionCenter(ActorContext<Message> context, ActorRef<AddressBook.Message> addressBook) {
        super(context);
        this.addressBook = addressBook;

        // Erhalten Sie die replyTo-Referenz aus dem Kontext
        ActorRef<List<ActorRef<Customer.Message>>> replyTo = context.spawnAnonymous(Behaviors.<List<ActorRef<Customer.Message>>>receive((ctx, message) -> {
            if (message instanceof List) {
                @SuppressWarnings("unchecked")
                List<ActorRef<Customer.Message>> customers = (List<ActorRef<Customer.Message>>) message;
                ctx.getLog().info("Kundenliste erhalten: {}", customers);
                // Nachricht senden, um Routen zu generieren
                context.getSelf().tell(new GenerateRoutes(customers));
                return Behaviors.same();
            }
            return Behaviors.unhandled();
        }));

        // Anfrage an das AddressBook senden, um die Kundenliste abzurufen
        addressBook.tell(new AddressBook.GetCustomers(replyTo));
    }

    /**
     * Factory-Methode zur Erstellung eines neuen Verteilzentrums-Akteurs.
     *
     * @param addressBook Referenz auf das Adressbuch
     * @return das Verhalten des Verteilzentrums-Akteurs
     */
    public static Behavior<Message> create(ActorRef<AddressBook.Message> addressBook) {
        return Behaviors.setup(context -> new DistributionCenter(context, addressBook));
    }
    // Definition der Nachrichten für das Verteilzentrum
    interface Message {}

    /**
     * Nachricht, die ankommende Pakete vom Lieferwagen enthält.
     */
    public static class ArriveMessage implements Message {
        private final List<Paket> pakete;
        private final ActorRef<DeliveryCar.Message> truck;

        public ArriveMessage(List<Paket> pakete, ActorRef<DeliveryCar.Message> truck) {
            this.pakete = pakete;
            this.truck = truck;
        }

        public List<Paket> getPakete() {
            return pakete;
        }

        public ActorRef<DeliveryCar.Message> getTruck() {
            return truck;
        }
    }

    /**
     * Nachricht zur Generierung von Routen für Lieferwagen.
     */
    public static class GenerateRoutes implements Message {
        private final List<ActorRef<Customer.Message>> customers;

        public GenerateRoutes(List<ActorRef<Customer.Message>> customers) {
            this.customers = customers;
        }

        public List<ActorRef<Customer.Message>> getCustomers() {
            return customers;
        }
    }

    @Override
    public Receive<Message> createReceive() {
        return newReceiveBuilder()
                .onMessage(ArriveMessage.class, this::onArriveMessage)
                .onMessage(GenerateRoutes.class, this::onGenerateRoutes)
                .onMessage(StartAllRoutes.class, this::onStartAllRoutes)
                .build();
    }

    /**
     * Handhabt die Ankunftsnachricht vom Lieferwagen.
     *
     * @param msg die Ankunftsnachricht
     * @return das Verhalten des Verteilzentrums
     */
    private Behavior<Message> onArriveMessage(ArriveMessage msg) {
        // Füge alle Pakete aus der Ankunftsnachricht dem Lagerraum hinzu
        List<Paket> arrivedPackages = msg.getPakete();
        lagerraum.addAll(arrivedPackages);
        getContext().getLog().info("Angekommene Pakete: {} Anz Packeten in Lager: {}", arrivedPackages.size(), lagerraum.size());
        msg.getTruck().tell(new DeliveryCar.Clear());

        // Entferne zufällig 3 Pakete aus dem Lagerraum und sende sie dem angekommenen Wagen
        int numPackagesToSend = Math.min(3, lagerraum.size());
        List<Paket> packagesToSend = new ArrayList<>();
        for (int i = 0; i < numPackagesToSend; i++) {
            packagesToSend.add(lagerraum.remove(0));
        }

        msg.getTruck().tell(new DeliveryCar.LoadMessage(packagesToSend));
        msg.getTruck().tell(new DeliveryCar.StartRoute()); // Starte die Route
        // Ausgabe, wie viele Pakete nun im Verteilzentrum vorhanden sind
        getContext().getLog().info("Pakete im Verteilzentrum nach Laden: {}", lagerraum.size());
        return this;
    }

    /**
     * Handhabt die Nachricht zur Generierung von Routen.
     *
     * @param msg die Nachricht zur Generierung von Routen
     * @return das Verhalten des Verteilzentrums
     */
    private Behavior<Message> onGenerateRoutes(GenerateRoutes msg) {
        List<ActorRef<Customer.Message>> customers = msg.getCustomers();
        generateRoutes(customers);
        return this;
    }

    /**
     * Generiert Routen für Lieferwagen.
     *
     * @param customers die Liste der Kunden
     */
    private void generateRoutes(List<ActorRef<Customer.Message>> customers) {
        if (customers == null || customers.isEmpty()) {
            getContext().getLog().info("Kundenliste ist leer oder null. Keine Routen generiert.");
            return;
        }

        for (int i = 0; i < 4; i++) { // Erstelle vier Lieferwagen
            List<ActorRef<Customer.Message>> route = generateRandomRoute(customers);
            ActorRef<DeliveryCar.Message> deliveryCar = getContext().spawn(DeliveryCar.create(route), "deliveryCar" + (i + 1));
            deliveryCars.add(deliveryCar);

            StringBuilder routeStr = new StringBuilder();
            for (int j = 0; j < route.size(); j++) {
                routeStr.append(route.get(j).path().name());
                if (j < route.size() - 1) {
                    routeStr.append(", ");
                }
            }
            getContext().getLog().info("Lieferwagen {} erstellt mit Route: {}", i + 1, routeStr);
            deliveryCar.tell(new DeliveryCar.StartRoute());
        }
    }

    /**
     * Startet alle Routen für die Lieferwagen.
     *
     * @param msg die Nachricht zum Starten aller Routen
     * @return das Verhalten des Verteilzentrums
     */
    private Behavior<Message> onStartAllRoutes(StartAllRoutes msg) {
        for (ActorRef<DeliveryCar.Message> deliveryCar : deliveryCars) {
            deliveryCar.tell(new DeliveryCar.StartRoute());
        }
        return this;
    }

    /**
     * Generiert eine zufällige Route aus einer Liste von Kunden.
     *
     * @param customers die Liste der Kunden
     * @return eine zufällige Route
     */
    private List<ActorRef<Customer.Message>> generateRandomRoute(List<ActorRef<Customer.Message>> customers) {
        List<ActorRef<Customer.Message>> route = new ArrayList<>(customers);
        Collections.shuffle(route);
        return route;
    }
}
